from ..common import DatasetType, BalancedType, GenerationMethod
from sdv.metadata import SingleTableMetadata
from sdv.single_table import CTGANSynthesizer, GaussianCopulaSynthesizer, CopulaGANSynthesizer, TVAESynthesizer
from IPython.display import display
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
import logging

class Synthesizer:
    def __init__(self, data, generation_params):
        self._metadata_obj = SingleTableMetadata()
        self._metadata_obj.detect_from_dataframe(data)
        self.data = data
        self.generation_params = generation_params
        self.num_rows = generation_params.num_rows
        self.epochs = generation_params.epochs
        self.target_variable = generation_params.target_variable
        self._prepare_dataset()
    
    def _log_info(self, message):
        current_logging_level = logging.getLogger().getEffectiveLevel()
        logging.getLogger().setLevel(logging.INFO)
        logging.info(message)
        logging.getLogger().setLevel(current_logging_level)
    
    def _prepare_dataset(self):
        if (self.generation_params.target_variable is None):
            self._log_info("You did not specify the name of the target variable (target_variable), so the dataset will be clustered and the target_variable will be - \"cluster\".")
            self._add_cluster_label()
            self.target_variable="cluster"
            self.metadata.add_column(
                column_name='cluster',
                sdtype='categorical')
        if (not self.generation_params.with_missing_values):
            self._impute_missing_values(self.data)

    def _impute_missing_values(self, data):
        numeric_columns = data.select_dtypes(include=['int64', 'float64']).columns
        categorical_columns = data.select_dtypes(include=['object']).columns
        
        if numeric_columns.any() and data[numeric_columns].isna().any().any():
            numeric_imputer = SimpleImputer(strategy='mean')
            data[numeric_columns] = numeric_imputer.fit_transform(data[numeric_columns])
        
        if categorical_columns.any() and data[categorical_columns].isna().any().any():
            categorical_imputer = SimpleImputer(strategy='most_frequent')
            data[categorical_columns] = categorical_imputer.fit_transform(data[categorical_columns])
    
    def _find_optimal_number_of_clusters(self, data, max_clusters=20):
        wcss = []
        for n in range(1, max_clusters):
            kmeans = KMeans(n_clusters=n, init='k-means++', max_iter=300, n_init=10, random_state=0)
            kmeans.fit(data)
            wcss.append(kmeans.inertia_)
        
        distances = []
        for i in range(1, len(wcss)-1):
            d1 = (wcss[i+1] - wcss[i]) / (i+2 - (i+1))
            d2 = (wcss[i] - wcss[i-1]) / ((i+1) - i)
            d2d1 = d1 - d2
            distances.append(d2d1)
        
        return distances.index(max(distances)) + 2

    def _add_cluster_label(self):
        if "cluster" in self.data.columns:
            return

        data_copy = self.data.copy()
        
        self._impute_missing_values(data_copy)

        numeric_features = data_copy.select_dtypes(include=['int64', 'float64']).columns
        categorical_features = data_copy.select_dtypes(include=['object', 'category']).columns

        numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])
        categorical_transformer = Pipeline(steps=[('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)])

        data_preprocessed = preprocessor.fit_transform(data_copy)

        n_clusters = self._find_optimal_number_of_clusters(data_preprocessed)

        kmeans = KMeans(n_clusters=n_clusters, init='k-means++', max_iter=300, n_init=10, random_state=0)
        cluster_labels = kmeans.fit_predict(data_preprocessed)

        self.data['cluster'] = cluster_labels

    def _determine_dataset_type(self):
        numerical_cols = self.metadata.get_column_names(sdtype='numerical')
        categorical_cols = self.metadata.get_column_names(sdtype='categorical')
        num_total_cols = len(numerical_cols) + len(categorical_cols)
        if len(numerical_cols) / num_total_cols > 0.6:
            return DatasetType.NUMERICAL
        elif len(categorical_cols) / num_total_cols > 0.6:
            return DatasetType.CATEGORICAL
        else:
            return DatasetType.MIXED

    def _determine_target_variable_balance(self):
        class_distribution = self.data[self.target_variable].value_counts(normalize=True)
        n_classes = len(class_distribution)

        max_percentage = class_distribution.max() * 100

        balanced_percentage = 100 / n_classes

        if max_percentage <= 1.1 * balanced_percentage:
            return BalancedType.BALANCED
        elif max_percentage <= 1.4 * balanced_percentage:
            return BalancedType.WEAKLY_BALANCED
        else:
            return BalancedType.NOT_BALANCED
        
    def _determine_generation_method(self, dataset_type, dataset_balanced_type):
        if (self.generation_params.fast_result or self.data.shape[1] > 99):
            return GenerationMethod.GAUSSIAN_COPULA
        elif (self.data.shape[0] < 101):
            return GenerationMethod.COPULA_GAN
        elif (dataset_balanced_type is BalancedType.BALANCED):
            if (dataset_type is DatasetType.NUMERICAL or dataset_type is DatasetType.MIXED):
                return GenerationMethod.TVAE
            return GenerationMethod.CONDITIONAL_GAN
        elif (dataset_balanced_type is BalancedType.WEAKLY_BALANCED):
            if (dataset_type is DatasetType.NUMERICAL or dataset_type is DatasetType.MIXED):
                return GenerationMethod.TVAE
            return GenerationMethod.COPULA_GAN
        elif (dataset_balanced_type is BalancedType.NOT_BALANCED):
            if (dataset_type is DatasetType.NUMERICAL or dataset_type is DatasetType.CATEGORICAL):
                return GenerationMethod.GAUSSIAN_COPULA
            return GenerationMethod.TVAE

    def visualize_metadata(self):
        visualization = self.metadata.visualize()
        if visualization is not None:
            display(visualization)

        summarized_visualization = self.metadata.visualize('summarized')
        if summarized_visualization is not None:
            display(summarized_visualization)
 
    @classmethod
    def create(cls, data, generation_params):
        synthesizer = cls(data, generation_params)
        synthesizer.visualize_metadata()

        return synthesizer

    @property
    def metadata(self):
        return self._metadata_obj

    @metadata.setter
    def metadata(self, value):
        if isinstance(value, SingleTableMetadata):
            self._metadata_obj = value
        else:
            raise ValueError("Incorrect type for metadata.")
        
    def ctgan_generate(self):
        synthesizer = CTGANSynthesizer(self.metadata, epochs=self.epochs, verbose=False)
        synthesizer.fit(self.data)
        data_generated = synthesizer.sample(num_rows=self.num_rows)
        
        return data_generated
    
    def gaussian_copula_generate(self):
        synthesizer = GaussianCopulaSynthesizer(self.metadata)
        synthesizer.fit(self.data)

        data_generated = synthesizer.sample(num_rows=self.num_rows)

        return data_generated
    
    def copula_gan_generate(self):
        synthesizer = CopulaGANSynthesizer(self.metadata, epochs=self.epochs, verbose=False)
        synthesizer.fit(self.data)

        data_generated = synthesizer.sample(num_rows=self.num_rows)

        return data_generated
    
    def tvae_generate(self):
        synthesizer = TVAESynthesizer(self.metadata, epochs=self.epochs)
        synthesizer.fit(self.data)

        data_generated = synthesizer.sample(num_rows=self.num_rows)

        return data_generated
    
    def generate(self):
        dataset_type = self._determine_dataset_type()
        dataset_balanced_type = self._determine_target_variable_balance()
        self._log_info(f"The type of the dataset is: {dataset_balanced_type.value}-{dataset_type.value}.")
        generation_method = self._determine_generation_method(dataset_type, dataset_balanced_type)
        self._log_info(f"Dataset will be generated by {generation_method.value} method.")

        generation_methods = {
            GenerationMethod.CONDITIONAL_GAN: self.ctgan_generate,
            GenerationMethod.GAUSSIAN_COPULA: self.gaussian_copula_generate,
            GenerationMethod.COPULA_GAN: self.copula_gan_generate,
            GenerationMethod.TVAE: self.tvae_generate,
        }

        generate_func = generation_methods.get(generation_method)

        data_generated = generate_func()

        return data_generated


import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.cluster import KMeans
from sklearn.impute import SimpleImputer
from table_evaluator import TableEvaluator
from sdv.evaluation.single_table import get_column_plot, run_diagnostic, evaluate_quality
from IPython.display import display, Markdown

class QualityEvaluator:
    def __init__(self, data, data_generated, metadata, generation_params):
        self.data = data.copy()
        self.data_generated = data_generated.copy()
        self.metadata = metadata
        self.generation_params = generation_params
        self.target_variable = generation_params.target_variable if generation_params.target_variable is not None else "cluster"
        self._prepare_datasets()

    def _prepare_datasets(self):
        self._impute_missing_values(self.data)
        self._impute_missing_values(self.data_generated)
    
    def _impute_missing_values(self, df):
        numeric_columns = df.select_dtypes(include=['int64', 'float64']).columns
        categorical_columns = df.select_dtypes(include=['object']).columns
        
        if numeric_columns.any() and df[numeric_columns].isna().any().any():
            numeric_imputer = SimpleImputer(strategy='mean')
            df[numeric_columns] = numeric_imputer.fit_transform(df[numeric_columns])
        
        if categorical_columns.any() and df[categorical_columns].isna().any().any():
            categorical_imputer = SimpleImputer(strategy='most_frequent')
            df[categorical_columns] = categorical_imputer.fit_transform(df[categorical_columns])

    def _combine_datasets(self):    
        data = self.data.copy()
        data_generated = self.data_generated.copy()

        data['is_synthetic'] = 0
        data_generated['is_synthetic'] = 1
        combined_data = pd.concat([data, data_generated], axis=0).reset_index(drop=True)

        categorical_columns = combined_data.select_dtypes(include=['object']).columns

        ohe = OneHotEncoder(sparse_output=False, handle_unknown='ignore')
        encoded_data = ohe.fit_transform(combined_data[categorical_columns])
        encoded_df = pd.DataFrame(encoded_data, columns=ohe.get_feature_names_out(categorical_columns))

        combined_data = combined_data.drop(columns=categorical_columns).join(encoded_df)
        combined_data = combined_data.sample(frac=1, random_state=42).reset_index(drop=True)

        return combined_data

    def _propensity_score(self):
        combined_data = self._combine_datasets()

        X = combined_data.drop('is_synthetic', axis=1)
        y = combined_data['is_synthetic']

        X_train, _, y_train, _ = train_test_split(X, y, test_size=0.2, random_state=42)

        clf = RandomForestClassifier(random_state=42)
        clf.fit(X_train, y_train)

        y_proba = clf.predict_proba(X)[:, 1]

        pMSE = np.mean((y_proba - 0.5) ** 2)

        print(f"Propensity score: {pMSE}")

        return pMSE

    def _log_cluster_metric(self, m=20):
        combined_data = self._combine_datasets()
        numeric_columns = combined_data.select_dtypes(include=['int64', 'float64']).columns
        scaler = StandardScaler()
        combined_data[numeric_columns] = scaler.fit_transform(combined_data[numeric_columns])

        kmeans = KMeans(n_clusters=m, n_init=10, random_state=42)
        kmeans.fit(combined_data[numeric_columns])
        combined_labels = kmeans.labels_

        boundary_index = len(self.data)
        ni = np.bincount(combined_labels, minlength=m)
        niR = np.bincount(combined_labels[:boundary_index], minlength=m)

        c = boundary_index / len(combined_labels)

        sum_of_squares = 0
        for i in range(m):
            sum_of_squares += (niR[i] / np.clip(ni[i], 1, None) - c) ** 2

        log_cluster_metric = np.log(sum_of_squares / m)
        print(f"Log-Cluster Metric: {log_cluster_metric}")

        return log_cluster_metric
    
    def _prediction_accuracy(self):
        X_generated = pd.get_dummies(self.data_generated.drop(self.target_variable, axis=1))
        y_generated = self.data_generated[self.target_variable]

        X_real = pd.get_dummies(self.data.drop(self.target_variable, axis=1))
        y_real = self.data[self.target_variable]

        X_real = X_real.reindex(columns=X_generated.columns, fill_value=0)
        
        clf = RandomForestClassifier(random_state=42)
        clf.fit(X_generated, y_generated)

        real_predictions = clf.predict(X_real)

        accuracy = accuracy_score(y_real, real_predictions)

        print(f'Prediction accuracy: {accuracy}')
        return accuracy

    def evaluate_metrics(self):
        propensity_score = self._propensity_score()
        log_cluster_metric = self._log_cluster_metric()
        prediction_accuracy = self._prediction_accuracy()

        return (propensity_score, log_cluster_metric, prediction_accuracy)

    def get_visual_evaluation(self):
        all_cols_set = set(self.metadata.get_column_names())
        numerical_cols_set = set(self.metadata.get_column_names(sdtype='numerical'))
        
        cat_cols = list(all_cols_set - numerical_cols_set)

        table_evaluator = TableEvaluator(self.data, self.data_generated, cat_cols=cat_cols, verbose=False)
        table_evaluator.evaluate(target_col=self.target_variable)
        table_evaluator.visual_evaluation()

    def print_column_plot(self, column_name, word_to_add = ''):
        fig = get_column_plot(
            real_data=self.data,
            synthetic_data=self.data_generated,
            column_name=column_name,
            metadata=self.metadata
        )

        title = (fig.layout.title.text if fig.layout.title else '') + ' ' + word_to_add

        fig.update_layout(title_text = title)

        fig.show()

    def evaluate_data_validity_with_structure(self, verbose=True):
        return run_diagnostic(
            real_data=self.data,
            synthetic_data=self.data_generated,
            metadata=self.metadata,
            verbose=verbose
        )

    def get_quality_report(self, verbose=True):
        return evaluate_quality(
            self.data,
            self.data_generated,
            self.metadata,
            verbose=verbose
        )
    
    def _print_distribution_plots(self, quality_report):
        column_shapes = quality_report.get_details('Column Shapes')

        rows_to_drop = column_shapes[column_shapes['Column'] == "cluster"].index
        column_shapes.drop(rows_to_drop, inplace=True)

        sorted_df = column_shapes.sort_values('Score', ignore_index=True)

        min_score_col = sorted_df.loc[0, 'Column']
        min_score = sorted_df.loc[0, 'Score']
        
        median_index = len(sorted_df) // 2
        median_score_col = sorted_df.loc[median_index, 'Column']
        median_score = sorted_df.loc[median_index, 'Score']
        
        max_score_col = sorted_df.loc[len(sorted_df) - 1, 'Column']
        max_score = sorted_df.loc[len(sorted_df) - 1, 'Score']

        self.print_column_plot(max_score_col, word_to_add=f'(Best distribution column: {max_score * 100:.2f}%)')
        self.print_column_plot(median_score_col, word_to_add=f'(Representative distribution column: {median_score * 100:.2f}%)')
        self.print_column_plot(min_score_col, word_to_add=f'(Worst distribution column: {min_score * 100:.2f}%)')
    
    def evaluate(self, with_details = False):
        self.evaluate_metrics()

        diagnostic_report = self.evaluate_data_validity_with_structure(verbose=False)
        data_validity = diagnostic_report.get_score()
        print(f'Data validity: {data_validity * 100:.2f}%')

        quality_report = self.get_quality_report(verbose=False)
        sdv_score = quality_report.get_score()
        print(f'SDV score: {sdv_score * 100:.2f}%')

        self._print_distribution_plots(quality_report)

        if (with_details):
            display(Markdown('# Table evaluator visualization'))
            self.get_visual_evaluation()



from enum import Enum

class DatasetType(Enum):
    NUMERICAL = "Numerical"
    CATEGORICAL = "Categorical"
    MIXED = "Mixed"

class BalancedType(Enum):
    BALANCED = "Balanced"
    WEAKLY_BALANCED = "Weakly Balanced"
    NOT_BALANCED = "Not Balanced"

class GenerationMethod(Enum):
    CONDITIONAL_GAN = "Conditional GAN"
    GAUSSIAN_COPULA = "Gaussian Copula"
    COPULA_GAN = "Copula GAN"
    TVAE = "Tabular VAE"

class GenerationParams:
    def __init__(self, num_rows, fast_result=False, target_variable=None, with_missing_values=True, epochs=1000):
        self.num_rows = num_rows
        self.fast_result=fast_result
        self.target_variable = target_variable
        self.with_missing_values = with_missing_values
        self.epochs = epochs